#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

#include "lib.h"
#include "iwram.h"
#include "irq.h"
#include "syscall.h"

#include "resource.h"
#include "resource\BGFont.h"
//#include "resource\BitMapFont.h"

//---------------------------------------------------------------------------
//Xe[W̃TCY
#define STAGE_CX            30
#define STAGE_CY            20

//Z̏
#define CELL_TYPE_DEAD      1
#define CELL_TYPE_ALIVE     2

//ZBG^C
#define BG_TILE_DEAD        0
#define BG_TILE_ALIVE       1

//Z̐
#define CELL_TIME           0x1000

//Q[̐is
#define GAME_STATE_INIT     1	//Xe[W̏
#define GAME_STATE_INPUT    2	//Z̍쐬
#define GAME_STATE_START    3	//Z̐Jn
#define GAME_STATE_SPLIT    4	//Z̐sȂ

//̓L[̃EFCg
#define KEY_RESPONSE        1

#define MAX(a,b) (((a)>(b))?(a):(b))
#define MIN(a,b) (((a)>(b))?(b):(a))
//---------------------------------------------------------------------------
//J[\
typedef struct {
	s16 x;
	s16 y;
	u16 cx;
	u16 cy;
	s16 mx;				//ړ
	s16 my;	
	s16 MinX;			//ړ
	s16 MinY;
	s16 MaxX;
	s16 MaxY;
	u16 SprNo;			//XvCg
	u16 SprTile;
	u16 SprSize;
	u16 SprForm;
	u8  key;			//ړp
} ST_CURSOR;

ST_CURSOR CURSOR;

//---------------------------------------------------------------------------
u8  BG2Chr;
u8  BG2Map;
u8  BG2Col;

u8  BG3Chr;
u8  BG3Map;
u8  BG3Col;

u8  Stage[STAGE_CX][STAGE_CY];
u8  ProgressTime;		//Žoߎ
u8  SplitTimingFlag;	//Z̐tOi^C}[荞݂ɂĎgpj

u8  GameStateFlag;		//Q[̐is
u16 KeyCount;			//L[EFCgp
u16 KeyOld;				//L[EFCgp
//---------------------------------------------------------------------------
void Init();
void InitCursor();

void GameStateInit();
void GameStateInput();
void GameStateStart();
void GameStateSplit();
void GameStatePause();

void DrawStageAll();
void DrawStage(const u8 x, const u8 y, const u8 type);
void DrawSprCursor();
void DrawBGProgressTime();

void Cursor();
void MoveCursor(ST_CURSOR* p);
u8   ChkMoveCursor(ST_CURSOR* p);

void SetNextStage();
void SetCell(const u8 x, const u8 y);
u8   GetCellAround(const s16 x, const s16 y);

void IRQUserSetup();
void IRQUserHandler();

void TimerStart();
void TimerStop();
void TimerReset();
//---------------------------------------------------------------------------
int main(void)
{
	Init();
	InitCursor();
	SetMode(MODE_0 | BG2_ENABLE | BG3_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);

	for(;;)
	{
		WaitForVsync();

		switch(GameStateFlag)
		{
			case GAME_STATE_INIT:   GameStateInit();   break;
			case GAME_STATE_INPUT:  GameStateInput();  break;
			case GAME_STATE_START:  GameStateStart();  break;
			case GAME_STATE_SPLIT:  GameStateSplit();  break;
			default: DebugMessage((u8*)"Err: Main GameStateFlag"); break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void Init()
{
	//EFCgRg[
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;

	//^C}[荞
	IRQUserSetup();

	//L[
	KeyInit();
	KeyCount = 0;
	KeyOld   = 0;

	//BG
	//BG2 \p
	//BG3 Q[p
	BG2Chr = 1;
	BG2Map = 0;
	BG2Col = 0;

	BG3Chr = 2;
	BG3Map = 1;
	BG3Col = 1;

	REG_BG2CNT = (BG_SIZEA_256_256    | BG_COLOR_16        |
				  BG_CHARBASE(BG2Chr) | BG_MAPBASE(BG2Map) | 0);

	REG_BG3CNT = (BG_SIZEA_256_256    | BG_COLOR_16        |
				  BG_CHARBASE(BG3Chr) | BG_MAPBASE(BG3Map) | 0);

	BGSetPal16  (BG2Col, (u16*)&BGFont_palette);
	BGSetChrData(BG2Chr, (u16*)&BGFont_data, BGFont_DATA_SIZE);

	BGSetPal16  (BG3Col, (u16*)&game_bg_pal_bin);
	BGSetChrData(BG3Chr, (u16*)&game_bg_img_bin, 2048);
	BGSetPalMap (BG3Map, 2048, 1);

	//Sprite
	SpriteInit();
	SpriteSetPal16((u16*)&game_spr_pal_bin);
	SpriteSetData ((u16*)&game_spr_img_bin, 2048);

	//Q[̐ݒ
	GameStateFlag = GAME_STATE_INIT;
}
//---------------------------------------------------------------------------
void InitCursor()
{
	CURSOR.x       = 0;
	CURSOR.y       = 0;
	CURSOR.cx      = 8;
	CURSOR.cy      = 8;
	CURSOR.mx      = CURSOR.cx;		//ړʂ̐ݒ
	CURSOR.my      = CURSOR.cy;
	CURSOR.MinX    = 0;				//ړ̐ݒ
	CURSOR.MaxX    = SCREEN_CX;
	CURSOR.MinY    = 0;
	CURSOR.MaxY    = SCREEN_CY;
	CURSOR.key     = 0x00;

	CURSOR.SprNo   = 0;
	CURSOR.SprTile = 0;
	CURSOR.SprSize = SP_SIZE_8;
	CURSOR.SprForm = SP_SQUARE;

	SpriteSetSize(CURSOR.SprNo, CURSOR.SprSize, CURSOR.SprForm, SP_COLOR_16);
	SpriteSetChr (CURSOR.SprNo, CURSOR.SprTile);
	SpriteMove   (CURSOR.SprNo, CURSOR.x, CURSOR.y);
}
//---------------------------------------------------------------------------
void GameStateInit()
{
	u8 x,y;

	//Xe[W̏
	for(y=0; y<STAGE_CY; y++)
	{
		for(x=0; x<STAGE_CX; x++)
		{
			Stage[x][y] = CELL_TYPE_DEAD;
		}
	}
	DrawStageAll();

	CURSOR.x = 0;
	CURSOR.y = 0;
	DrawSprCursor();

	ProgressTime = 0;
	DrawBGProgressTime();

	GameStateFlag = GAME_STATE_INPUT;
}
//---------------------------------------------------------------------------
void DrawStageAll()
{
	u8 x,y;

	for(y=0; y<STAGE_CY; y++)
	{
		for(x=0; x<STAGE_CX; x++)
		{
			DrawStage(x, y, Stage[x][y]);
		}
	}
}
//---------------------------------------------------------------------------
void DrawStage(const u8 x, const u8 y, const u8 type)
{
	u8 tileNo = 0x00;
	switch(type)
	{
		case CELL_TYPE_ALIVE: tileNo = BG_TILE_ALIVE; break;
		case CELL_TYPE_DEAD:  tileNo = BG_TILE_DEAD;  break;
		default: DebugMessage((u8*)"Err: DrawStage type"); break;
	}

	BGDrawChar(BG3Map, x, y, BG3Col, tileNo);
}
//---------------------------------------------------------------------------
void GameStateInput()
{
	Cursor();
}
//---------------------------------------------------------------------------
void IRQUserSetup()
{
	REG_IME = IRQ_MASTER_OFF;

	REG_INTERUPT  = (u32)IRQUserHandler;
	REG_IE       |= IRQ_BIT_TIMER0;

	REG_IME = IRQ_MASTER_ON;
}
//---------------------------------------------------------------------------
void IRQUserHandler()
{
	REG_IME  = IRQ_MASTER_OFF;
	u16 Flag = REG_IF;

	if(Flag & IRQ_BIT_TIMER0)
	{
		//oߎԂi߂tOnmɂ܂
		SplitTimingFlag = TRUE;
	}

	REG_IF  = Flag;
	REG_IME = IRQ_MASTER_ON;
}
//---------------------------------------------------------------------------
void TimerStart()
{
	SplitTimingFlag = FALSE;
	REG_TM0D   = CELL_TIME;
	REG_TM0CNT = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;

}
//---------------------------------------------------------------------------
void TimerStop()
{
	REG_TM0CNT = TM_DISABLE;
	REG_TM0D   = 0x0000;
}
//---------------------------------------------------------------------------
void TimerReset()
{
	REG_TM0D = CELL_TIME;
	SplitTimingFlag = FALSE;
}
//---------------------------------------------------------------------------
void Cursor()
{
	u8 Key = KeyInput();

	//L[X|X(TODO ꃉCuj
	if(Key == 0x00 && KeyOld != Key)
	{
		KeyCount = 0;
		KeyOld   = Key;
		return;
	}
	if(KeyCount++ <= KEY_RESPONSE)
	{
		return;
	}

	KeyCount = 0;
	CURSOR.key = 0x00;

	if(Key & KEY_UP   ) CURSOR.key |= KEY_UP;
	if(Key & KEY_DOWN ) CURSOR.key |= KEY_DOWN;
	if(Key & KEY_LEFT ) CURSOR.key |= KEY_LEFT;
	if(Key & KEY_RIGHT) CURSOR.key |= KEY_RIGHT;
	if(Key & KEY_A    ) CURSOR.key |= KEY_A;
	if(Key & KEY_START) CURSOR.key |= KEY_START;

	//ړ`FbN
	if( ChkMoveCursor(&CURSOR) == TRUE )
	{
		MoveCursor(&CURSOR);
		DrawSprCursor();
	}
	//Z̐ݒu
	if(CURSOR.key & KEY_A)
	{
		SetCell(CURSOR.x/CURSOR.mx, CURSOR.y/CURSOR.my);
	}
	//Z̐X^[g
	if(CURSOR.key & KEY_START)
	{
		GameStateFlag = GAME_STATE_START;
	}
}
//---------------------------------------------------------------------------
//ړ͈ړ\`FbN܂
u8 ChkMoveCursor(ST_CURSOR* p)
{
	s16 tx1 = p->x;
	s16 ty1 = p->y;

	if(p->key & KEY_UP   ) ty1-=p->my;
	if(p->key & KEY_DOWN ) ty1+=p->my;
	if(p->key & KEY_LEFT ) tx1-=p->mx;
	if(p->key & KEY_RIGHT) tx1+=p->mx;

	if(tx1< CURSOR.MinX || ty1< CURSOR.MinY) return FALSE;
	if(tx1>=CURSOR.MaxX || ty1>=CURSOR.MaxY) return FALSE;

	return TRUE;
}
//---------------------------------------------------------------------------
//ړ
void MoveCursor(ST_CURSOR* p)
{
	if(p->key & KEY_UP   ) p->y-=p->my;
	if(p->key & KEY_DOWN ) p->y+=p->my;
	if(p->key & KEY_LEFT ) p->x-=p->mx;
	if(p->key & KEY_RIGHT) p->x+=p->mx;
}
//---------------------------------------------------------------------------
void DrawSprCursor()
{
	SpriteMove(CURSOR.SprNo, CURSOR.x, CURSOR.y);
}
//---------------------------------------------------------------------------
void GameStateStart()
{
	//J[\ʊO
	CURSOR.x = SCREEN_CX;
	CURSOR.y = SCREEN_CY;
	DrawSprCursor();

	TimerStart();
	GameStateFlag = GAME_STATE_SPLIT;
}
//---------------------------------------------------------------------------
//Z̐
void GameStateSplit()
{
	//tÕ`FbN
	if(SplitTimingFlag == FALSE)
	{
		return;
	}
	//Xe[W̍XV
	SetNextStage();
	DrawStageAll();

	//oߎԂ̕\
	ProgressTime++;
	DrawBGProgressTime();

	TimerReset();

	u8 Key = KeyInput();
	if(Key & KEY_B)
	{
		//𒆒fāAXe[W̏֏ԂJڂ܂
		GameStateFlag = GAME_STATE_INIT;
	}
}
//---------------------------------------------------------------------------
//Z̐ݒu
void SetCell(const u8 x, const u8 y)
{
	u8 type = Stage[x][y];

	if(type == CELL_TYPE_ALIVE)
	{
		Stage[x][y] = CELL_TYPE_DEAD;
	}
	else
	{
		Stage[x][y] = CELL_TYPE_ALIVE;
	}

	DrawStage(x,y,Stage[x][y]);
}
//---------------------------------------------------------------------------
//mx,myӂ̐ĂZJEg܂
u8 GetCellAround(const s16 mx, const s16 my)
{
	s16 x,y;
	u8  RetCnt = 0;

	for(y=MAX(my-1,0); y<=MIN(my+1,STAGE_CY-1); y++)
	{
		for(x=MAX(mx-1,0); x<=MIN(mx+1,STAGE_CX-1); x++)
		{
			if(mx == x && my == y)
			{
				continue;
			}
			if(Stage[x][y] == CELL_TYPE_ALIVE)
			{
				RetCnt++;
			}
		}
	}
	return RetCnt;
}
//---------------------------------------------------------------------------
void DrawBGProgressTime()
{
	char msg[40];
	bprintf(msg, "%3d", ProgressTime);
	BGDrawStr(BG2Map, 25, 18, (u8*)msg);
}
//---------------------------------------------------------------------------
void SetNextStage()
{
	u8 tmpStage[STAGE_CX][STAGE_CY];
	u8 x,y;
	u8 type;

	for(y=0; y<STAGE_CY; y++)
	{
		for(x=0; x<STAGE_CX; x++)
		{
			switch( GetCellAround(x,y) )
			{
				case 2:  type = Stage[x][y];     break;
				case 3:  type = CELL_TYPE_ALIVE; break;
				default: type = CELL_TYPE_DEAD;  break;
			}
			tmpStage[x][y] = type;
		}
	}

	for(y=0; y<STAGE_CY; y++)
	{
		for(x=0; x<STAGE_CX; x++)
		{
			Stage[x][y] = tmpStage[x][y];
		}
	}
}
